#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <arpa/inet.h>
#include <curl/curl.h>
#include <sys/socket.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/epoll.h>
#include <signal.h>
#include <errno.h>
#include <ctype.h>
#include <vector>
#include <string>
#include <map>

#include "metacube.h"
#include "server.h"
#include "input.h"
#include "state.pb.h"

#define NUM_SERVERS 4
#define STREAM_ID "stream"
#define STREAM_URL "http://gruessi.zrh.sesse.net:4013/"

using namespace std;

Server *servers = NULL;
volatile bool hupped = false;

void hup(int ignored)
{
	hupped = true;
}

int create_server_socket(int port)
{
	int server_sock = socket(PF_INET6, SOCK_STREAM, IPPROTO_TCP);
	if (server_sock == -1) {
		perror("socket");
		exit(1);
	}

	int one = 1;
	if (setsockopt(server_sock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)) == -1) {
		perror("setsockopt(SO_REUSEADDR)");
		exit(1);
	}

	// We want dual-stack sockets. (Sorry, OpenBSD and Windows XP...)
	int zero = 0;
	if (setsockopt(server_sock, IPPROTO_IPV6, IPV6_V6ONLY, &zero, sizeof(zero)) == -1) {
		perror("setsockopt(IPV6_V6ONLY)");
		exit(1);
	}

	sockaddr_in6 addr;
	memset(&addr, 0, sizeof(addr));
	addr.sin6_family = AF_INET6;
	addr.sin6_port = htons(port);

	if (bind(server_sock, reinterpret_cast<sockaddr *>(&addr), sizeof(addr)) == -1) {
		perror("bind");
		exit(1);
	}

	if (listen(server_sock, 128) == -1) {
		perror("listen");
		exit(1);
	}

	return server_sock;
}

void *acceptor_thread_run(void *arg)
{
	int server_sock = int(intptr_t(arg));
	int num_accepted = 0;
	for ( ;; ) {
		sockaddr_in6 addr;
		socklen_t addrlen = sizeof(addr);

		// Get a new socket.
		int sock = accept(server_sock, reinterpret_cast<sockaddr *>(&addr), &addrlen);
		if (sock == -1 && errno == EINTR) {
			continue;
		}
		if (sock == -1) {
			perror("accept");
			exit(1);
		}

		// Set the socket as nonblocking.
		int one = 1;
		if (ioctl(sock, FIONBIO, &one) == -1) {
			perror("FIONBIO");
			exit(1);
		}

		// Pick a server, round-robin, and hand over the socket to it.
		servers[num_accepted % NUM_SERVERS].add_client(sock);
		++num_accepted;	
	}
}

// Serialize the given state to a file descriptor, and return the (still open)
// descriptor.
int make_tempfile(const CubemapStateProto &state)
{
	char tmpl[] = "/tmp/cubemapstate.XXXXXX";
	int state_fd = mkstemp(tmpl);
	if (state_fd == -1) {
		perror("mkstemp");
		exit(1);
	}

	string serialized;
	state.SerializeToString(&serialized);

	const char *ptr = serialized.data();
	size_t to_write = serialized.size();
	while (to_write > 0) {
		ssize_t ret = write(state_fd, ptr, to_write);
		if (ret == -1) {
			perror("write");
			exit(1);
		}

		ptr += ret;
		to_write -= ret;
	}

	return state_fd;
}

// Read the state back from the file descriptor made by make_tempfile,
// and close it.
CubemapStateProto read_tempfile(int state_fd)
{
	if (lseek(state_fd, 0, SEEK_SET) == -1) {
		perror("lseek");
		exit(1);
	}

	string serialized;
	char buf[4096];
	for ( ;; ) {
		ssize_t ret = read(state_fd, buf, sizeof(buf));
		if (ret == -1) {
			perror("read");
			exit(1);
		}
		if (ret == 0) {
			// EOF.
			break;
		}

		serialized.append(string(buf, buf + ret));
	}

	close(state_fd);  // Implicitly deletes the file.

	CubemapStateProto state;
	if (!state.ParseFromString(serialized)) {
		fprintf(stderr, "PANIC: Failed deserialization of state.\n");
		exit(1);
	}

	return state;
}

// Split a line on whitespace, e.g. "foo  bar baz" -> {"foo", "bar", "baz"}.
vector<string> split_tokens(const string &line)
{
	vector<string> ret;
	string current_token;

	for (size_t i = 0; i < line.size(); ++i) {
		if (isspace(line[i])) {
			if (!current_token.empty()) {
				ret.push_back(current_token);
			}
			current_token.clear();
		} else {
			current_token.push_back(line[i]);
		}
	}
	if (!current_token.empty()) {
		ret.push_back(current_token);
	}
	return ret;
}

struct ConfigLine {
	string keyword;
	vector<string> arguments;
	map<string, string> parameters;
};

// Parse the configuration file.
vector<ConfigLine> parse_config(const string &filename)
{
	vector<ConfigLine> ret;

	FILE *fp = fopen(filename.c_str(), "r");
	if (fp == NULL) {
		perror(filename.c_str());
		exit(1);
	}

	char buf[4096];
	while (!feof(fp)) {
		if (fgets(buf, sizeof(buf), fp) == NULL) {
			break;
		}

		// Chop off the string at the first #, \r or \n.
		buf[strcspn(buf, "#\r\n")] = 0;

		// Remove all whitespace from the end of the string.
		size_t len = strlen(buf);
		while (len > 0 && isspace(buf[len - 1])) {
			buf[--len] = 0;
		}

		// If the line is now all blank, ignore it.
		if (len == 0) {
			continue;
		}

		vector<string> tokens = split_tokens(buf);
		assert(!tokens.empty());
		
		ConfigLine line;
		line.keyword = tokens[0];

		for (size_t i = 1; i < tokens.size(); ++i) {
			// foo=bar is a parameter; anything else is an argument.
			size_t equals_pos = tokens[i].find_first_of('=');
			if (equals_pos == string::npos) {
				line.arguments.push_back(tokens[i]);
			} else {
				string key = tokens[i].substr(0, equals_pos);
				string value = tokens[i].substr(equals_pos + 1, string::npos);
				line.parameters.insert(make_pair(key, value));
			}
		}

		ret.push_back(line);
	}

	fclose(fp);
	return ret;
}

int main(int argc, char **argv)
{
	fprintf(stderr, "\nCubemap starting.\n");

	string config_filename = (argc == 1) ? "cubemap.config" : argv[1];
	vector<ConfigLine> config = parse_config(config_filename);

	// Go through each (parsed) configuration line.
	int port = -1;
	for (unsigned i = 0; i < config.size(); ++i) {
		if (config[i].keyword == "port") {
			if (config[i].parameters.size() > 0 ||
			    config[i].arguments.size() != 1) {
				fprintf(stderr, "ERROR: 'port' takes one argument and no parameters\n");
				exit(1);
			}
			port = atoi(config[i].arguments[0].c_str());
		}
	}
	if (port <= 0 || port > 65535) {
		fprintf(stderr, "ERROR: Missing or invalid 'port' statement in config file\n");
		exit(1);
	}

	// Create the servers.
	servers = new Server[NUM_SERVERS];

	int server_sock = -1, old_port = -1;
	if (argc == 4 && strcmp(argv[2], "-state") == 0) {
		fprintf(stderr, "Deserializing state from previous process... ");
		int state_fd = atoi(argv[3]);
		CubemapStateProto loaded_state = read_tempfile(state_fd);

		// Deserialize the streams.
		for (int i = 0; i < loaded_state.streams_size(); ++i) {
			for (int j = 0; j < NUM_SERVERS; ++j) {
				servers[j].add_stream_from_serialized(loaded_state.streams(i));
			}
		}

		// Put back the existing clients. It doesn't matter which server we
		// allocate them to, so just do round-robin.
		for (int i = 0; i < loaded_state.clients_size(); ++i) {
			servers[i % NUM_SERVERS].add_client_from_serialized(loaded_state.clients(i));
		}

		// Deserialize the server socket.
		server_sock = loaded_state.server_sock();
		old_port = loaded_state.port();

		fprintf(stderr, "done.\n");
	} else{
		// TODO: This should come from the config file.
		for (int i = 0; i < NUM_SERVERS; ++i) {
			servers[i].add_stream(STREAM_ID);
		}
	}

	// Open a new server socket if we do not already have one, or if we changed ports.
	if (server_sock != -1 && port != old_port) {
		fprintf(stderr, "NOTE: Port changed from %d to %d; opening new socket.\n", old_port, port);
		close(server_sock);
		server_sock = -1;
	}
	if (server_sock == -1) {
		server_sock = create_server_socket(port);
	}

	// Start up all the servers!
	for (int i = 0; i < NUM_SERVERS; ++i) {
		servers[i].run();
	}

	pthread_t acceptor_thread;
	pthread_create(&acceptor_thread, NULL, acceptor_thread_run, reinterpret_cast<void *>(server_sock));

	// TODO: This should come from the config file.
	Input input(STREAM_ID, STREAM_URL);
	input.run();

	signal(SIGHUP, hup);

	while (!hupped) {
		usleep(100000);
	}

	input.stop();

	CubemapStateProto state;
	state.set_server_sock(server_sock);
	state.set_port(port);
	for (int i = 0; i < NUM_SERVERS; ++i) {
		servers[i].stop();

		CubemapStateProto local_state = servers[i].serialize();

		// The stream state should be identical between the servers, so we only store it once.
		if (i == 0) {
			state.mutable_streams()->MergeFrom(local_state.streams());
		}
		for (int j = 0; j < local_state.clients_size(); ++j) {
			state.add_clients()->MergeFrom(local_state.clients(j));
		}
	}
	delete[] servers;

	fprintf(stderr, "Serializing state and re-execing...\n");
	int state_fd = make_tempfile(state);

	char buf[16];
	sprintf(buf, "%d", state_fd);

	for ( ;; ) {
		execlp(argv[0], argv[0], config_filename.c_str(), "-state", buf, NULL);
		perror("execlp");
		fprintf(stderr, "PANIC: re-exec of %s failed. Waiting 0.2 seconds and trying again...\n", argv[0]);
		usleep(200000);
	}
}
