#ifndef _SERVER_H
#define _SERVER_H 1

#include <stdint.h>
#include <pthread.h>
#include <string>
#include <map>

#define NUM_SERVERS 4
#define BACKLOG_SIZE 1048576
#define EPOLL_MAX_EVENTS 8192
#define EPOLL_TIMEOUT_MS 20
#define MAX_CLIENT_REQUEST 16384

class ClientProto;
class CubemapStateProto;
class StreamProto;

struct Client {
	Client() {}
	Client(int sock);

	// Serialization/deserialization.
	Client(const ClientProto &serialized);
	ClientProto serialize() const;

	// The file descriptor associated with this socket.
	int sock;

	enum State { READING_REQUEST, SENDING_HEADER, SENDING_DATA };
	State state;

	// The HTTP request, as sent by the client. If we are in READING_REQUEST,
	// this might not be finished.
	std::string request;

	// What stream we're connecting to; parsed from <request>.
	// Not relevant for READING_REQUEST.
	std::string stream_id;

	// The header we want to send. This is nominally a copy of Stream::header,
	// but since that might change on reconnects etc., we keep a local copy here.
	// Only relevant for SENDING_HEADER; blank otherwise.
	std::string header;

	// Number of bytes we've sent of the header. Only relevant for SENDING_HEADER.
	size_t header_bytes_sent;

	// Number of bytes we've sent of data. Only relevant for SENDING_DATA.
	size_t bytes_sent;
};

struct Stream {
	Stream(const std::string &stream_id);
	~Stream();

	// Serialization/deserialization.
	Stream(const StreamProto &serialized);
	StreamProto serialize() const;

	std::string stream_id;

	// The HTTP response header, plus the video stream header (if any).
	std::string header;

	// The stream data itself, stored in a circular buffer.
	char *data;

	// How many bytes <data> contains. Can very well be larger than BACKLOG_SIZE,
	// since the buffer wraps.
	size_t data_size;

private:
	Stream(const Stream& other);
};

class Server {
public:
	Server();
	~Server();

	// Start a new thread that handles clients.
	void run();

	// Stop the thread.
	void stop();

	CubemapStateProto serialize() const;

	void add_client(int sock);
	void add_client_from_serialized(const ClientProto &client);

	void add_stream(const std::string &stream_id);
	void add_stream_from_serialized(const StreamProto &stream);

	void set_header(const std::string &stream_id, const std::string &header);
	void add_data(const std::string &stream_id, const char *data, size_t bytes);

private:
	pthread_t worker_thread;

	// All variables below this line are protected by the mutex.
	pthread_mutex_t mutex;

	// If the thread should stop or not.
	bool should_stop;	

	// Map from stream ID to stream.
	std::map<std::string, Stream *> streams;

	// Map from file descriptor to client.
	std::map<int, Client> clients;

	// Used for epoll implementation (obviously).
	int epoll_fd;
	epoll_event events[EPOLL_MAX_EVENTS];

	// Clients that are in SENDING_DATA, but that we don't listen on,
	// because we currently don't have any data for them.
	// See put_client_to_sleep() and wake_up_all_clients().
	std::vector<int> sleeping_clients;

	// Recover the this pointer, and call do_work().
	static void *do_work_thunk(void *arg);

	// The actual worker thread.
	void do_work();

	void process_client(Client *client);

	// Close a given client socket, and clean up after it.
	void close_client(Client *client);

	// Parse the HTTP request.
	void parse_request(Client *client);

	// Construct the HTTP header, and set the client into
	// the SENDING_HEADER state.
	void construct_header(Client *client);

	// Put client to sleep, since there is no more data for it; we will on
	// longer listen on POLLOUT until we get more data. Also, it will be put
	// in the list of clients to wake up when we do.
	void put_client_to_sleep(Client *client);

	// We have more data, so mark all clients that are sleeping as ready to go.
	void wake_up_all_clients();

	// TODO: This function should probably die.
	Stream *find_stream(const std::string &stream_id);
};

#endif  // !defined(_SERVER_H)
